using System;
using System.Xml;
using System.Xml.Xsl;
using System.Xml.XPath;
using System.IO;


public class Test
{	static void Main(string[] paras)
	{	POSD posd = new POSD(paras[0]);
	}
}





public class POSD
{	SVG svg;
	
	public POSD(string paras)
	{	Parameters parameters = new Parameters(paras);
//		parameters.diagnose("temporary" + parameters.get("entry") + ".xml");
		Database database = new Database(parameters);
		svg = new SVG(database, parameters);
		parameters.release();
	}

	public XmlElement get()
	{	return svg.get();
	}
}





internal class SVG
{	XmlElement output;

	internal SVG(Database database, Parameters parameters)
	{	XmlElement input = formInput(database, parameters);	
//		parameters.diagnose(input, "tempin" + parameters.get("entry") + ".xml");
		output = transformInput(new XPathDocument(new XmlNodeReader(input)), parameters);
//		parameters.diagnose(output, "tempout" + parameters.get("entry") + ".xml");
	}
	
	internal XmlElement get()
	{	return output;
	}

	private XmlElement formInput(Database database, Parameters parameters)
	{	XmlDocument input = parameters.load();
		XmlNodeList net = database.getLevels(parameters.get("current"), parameters.get(parameters.get("style")+"search"));  
		for (int i = 0; i < net.Count; i++)	input.DocumentElement.AppendChild(input.CreateElement("augusta"));
		XmlNodeList list = input.DocumentElement.SelectNodes("augusta");	
		for (int i = 0; i < net.Count; i++) list[i].InnerXml = net[i].InnerXml;
		return input.DocumentElement;
	}

	private XmlElement transformInput(XPathDocument dummy, Parameters parameters)
	{	XslTransform transformation = new XslTransform();
		transformation.Load(parameters.get("physical") + "\\posd.xsl");
		XsltArgumentList args = null;
		XmlResolver resolver = null;
		XmlReader reader = transformation.Transform(dummy, args, resolver);
		XmlDocument document = new XmlDocument();
		document.Load(reader);
		try {document.Save(parameters.get("physical") + "\\posd.svg");}
		finally {}
		return (XmlElement) document.DocumentElement;
	}
}





internal class Database
{	Parameters parameters;

	internal Database(Parameters paras) 
	{	parameters = paras;
		processActionParameters();
	}

	internal XmlNode get(string identifier)
	{	XmlDocument input = new XmlDocument();
		input.Load(parameters.get("dblink") + "\\" + identifier + ".xml");
		return input.DocumentElement;
	}

	internal void put(string identifier, XmlElement content)
	{	content.OwnerDocument.Save(parameters.get("dblink") + "\\" + identifier + ".xml");
	}

	internal XmlNodeList getLevels(string identifier, string control)
	{	int[] con = formControl(control, new int[3]{1,0,0});
		parameters.put("level", "" + con[0]);
		if (XmlConvert.ToInt16(parameters.get("counter")) > 1000) parameters.put("counter", "0");
		else parameters.put("counter", "" + (XmlConvert.ToInt16(parameters.get("counter")) + 1));
		return fillLevels(insertStartNode(initialiseLevels(con[0]), identifier, con[1]), con[1], control);
	}

	private void processActionParameters()
	{	if (parameters.get("action") == "create") create();
		else if (parameters.get("action") == "attach") attach();
		else if (parameters.get("action") == "detach") detach();
		else if (parameters.get("action") == "rename") rename();
		else if (parameters.get("action") == "retype") redo("type");
		else if (parameters.get("action") == "reannotate") redo("annotation");
		else if (parameters.get("action") == "move") move();
		else if (parameters.get("action") == "link") link();
		parameters.put("action", "|");
		parameters.put("param2", "|");
	}
	
	private void create() 
	{	create(parameters.get("current"), parameters.get("param1"));
	}

	private void create(string composite, string name) 
	{	XmlDocument document = new XmlDocument();
		XmlNode hazel = document.AppendChild(document.CreateElement("hazel"));
		string component = getfree();
		hazel.AppendChild(document.CreateElement("identifier")).InnerText = component;
		hazel.AppendChild(document.CreateElement("type")).InnerText = "";
		hazel.AppendChild(document.CreateElement("annotation")).InnerText = "";
		document.Save(parameters.get("dblink") + "\\" + component + ".xml");
		attach(component, composite, name);
	}

	private string getfree()
	{	return "s" + System.IO.Directory.GetFiles(parameters.get("dblink"), "s*").Length;
	}
	
	private void attach()		
	{	attach(parameters.get("param1"), parameters.get("current"), parameters.get("param2"));
	}

	private void attach(string component, string composite, string name)		
	{	XmlDocument document = new XmlDocument();
		document.Load(parameters.get("dblink") + "\\" + composite + ".xml");
		if (document.DocumentElement.SelectSingleNode("component[identifier='" + component + "']") == null)
		{	XmlNode node = document.DocumentElement.AppendChild(document.CreateElement("component"));
			node.AppendChild(document.CreateElement("identifier")).InnerText = component; 
			node.AppendChild(document.CreateElement("name")).InnerText = name; 
			document.Save(parameters.get("dblink") + "\\" + composite + ".xml");
			document.Load(parameters.get("dblink") + "\\" + component + ".xml");
			document.DocumentElement.AppendChild(document.CreateElement("composite")).
				AppendChild(document.CreateElement("identifier")).InnerText = composite;
			document.Save(parameters.get("dblink") + "\\" + component + ".xml");
		}
	}
	
	private void detach()
	{	string current = parameters.get("current");
		string component = parameters.get("param1");
		XmlDocument document = new XmlDocument();
		document.Load(parameters.get("dblink") + "\\" + current + ".xml");
		XmlNode node = document.DocumentElement.SelectSingleNode("component[identifier='" + component + "']");
		if (node != null) document.DocumentElement.RemoveChild(node);
		document.Save(parameters.get("dblink") + "\\" + current + ".xml");
		document.Load(parameters.get("dblink") + "\\" + component + ".xml");
		XmlNode composite = document.DocumentElement.SelectSingleNode("composite[identifier='" + current + "']");
		if (composite != null) document.DocumentElement.RemoveChild(composite);
		document.Save(parameters.get("dblink") + "\\" + component + ".xml");
	}

	private void rename() 
	{	string current = parameters.get("current");
		string target = parameters.get("saved");
		if (parameters.get("style") != "graph") target = parameters.get("selected");
		if (current != "" && target != "" && current != target)
		{	XmlDocument document = new XmlDocument();
			document.Load(parameters.get("dblink") + "\\" + current + ".xml");
			XmlNode component = document.DocumentElement.SelectSingleNode("component[identifier='" + target + "']");
			if (component != null) 
			{	XmlNode targetNode = component.SelectSingleNode("name");
				if (targetNode == null) targetNode = component.AppendChild(document.CreateElement("name"));
				targetNode.InnerText = parameters.get("param1");
			}
			document.Save(parameters.get("dblink") + "\\" + current + ".xml");
		}
	}	

	private void redo(string typeorannotation) 
	{	string target = parameters.get("current");
		if (target != "")
		{	XmlDocument document = new XmlDocument();
			document.Load(parameters.get("dblink") + "\\" + target + ".xml");
			document.DocumentElement.SelectSingleNode(typeorannotation).InnerText = parameters.get("param1");
			document.Save(parameters.get("dblink") + "\\" + target + ".xml");
		}
	}	
		
	private void move()
	{	string[] paras = parameters.get("param1").Split('|');		
		parameters.put("selected", paras[1]);
		if (paras[2] == "")
		{	XmlDocument document = new XmlDocument();
			document.Load(parameters.get("dblink") + "\\" + paras[1] + ".xml");
			XmlNode picture = document.DocumentElement.SelectSingleNode(
				"position[type='" + paras[0] + "' and identifier='" + parameters.get("current") + "']");
			if (picture != null) 
			{	picture.SelectSingleNode("x").InnerText = paras[3];
				picture.SelectSingleNode("y").InnerText = paras[4];
			}
			else
			{	picture = document.DocumentElement.AppendChild(document.CreateElement("position"));
				picture.AppendChild(document.CreateElement("identifier")).InnerText = parameters.get("current");
				picture.AppendChild(document.CreateElement("type")).InnerText = paras[0];
				picture.AppendChild(document.CreateElement("x")).InnerText = paras[3];
				picture.AppendChild(document.CreateElement("y")).InnerText = paras[4];
			}
			document.Save(parameters.get("dblink") + "\\" + paras[1] + ".xml");
		}
	}

	private void link()
	{	string[] paras = parameters.get("param1").Split('|');	
		string name = parameters.get("param2");	
		parameters.put("selected", paras[1]);
		if (paras[1] != paras[2] && paras[2] != "")
		{	string component = getfree();
			create(paras[1], name);
			attach(component, paras[2], name);
		}
	}
	
	private int[] formControl(string control, int[] reply)
//		reply = no of levels, start level, current level	
	{	if (control == "e") return reply;
		if (((isNext(control, "p") || isNext(control, "d")) && reply[2] == reply[0] - 1)
				|| ((isNext(control, "m") || isNext(control, "u")) && reply[2] == 0))
			reply[0] = reply[0] + 1;
		if ((isNext(control, "m") || isNext(control, "u")) && reply[2] == 0) reply[1] = reply[1] + 1;
		if (isNext(control, "p") || (isNext(control, "u") && reply[2] == 0)) reply[2] = reply[2] + 1;
		else if (isNext(control, "m") && reply[2] > 0) reply[2] = reply[2] - 1;
		return formControl(control.Substring(1), reply);
	}
	
	private bool isNext(string control, string test) {	return control.Substring(0, 1) == test;} 

	private XmlNodeList initialiseLevels(int levels)
	{	XmlDocument output = new XmlDocument();
		XmlNode hazel = output.AppendChild(output.CreateElement("hazel"));
		for (int i = 0; i < levels; i++) hazel.AppendChild(output.CreateElement("augusta"));
		return hazel.ChildNodes;
	}

	private XmlNodeList insertStartNode(XmlNodeList list, string identifier, int level)
	{	XmlDocument input = new XmlDocument();
		input.Load(parameters.get("dblink") + "\\" + identifier + ".xml");			
		XmlNode alice = list[level].AppendChild(list[level].OwnerDocument.CreateElement("alice"));
		alice.InnerXml = input.DocumentElement.InnerXml;
		return list;
	}

	private XmlNodeList fillLevels(XmlNodeList levels, int level, string control)
	{	XmlDocument output = levels[0].OwnerDocument;
		if (control == "e") return levels;
		if (control.Substring(0, 1) == "p") return fillLevels(levels, level + 1, control.Substring(1)); 
		if (control.Substring(0, 1) == "m") return fillLevels(levels, level - 1, control.Substring(1)); 
		XmlNodeList nodes = levels[level].ChildNodes;
		for (int i = 0; i < nodes.Count; i++)
		{	if (control.Substring(0, 1) == "d") fillLevel(nodes[i], levels[level + 1], "component");	
			else fillLevel(nodes[i], levels[level - 1], "composite");	
		}
		return fillLevels(levels, level, control.Substring(1));
	}    

	private void fillLevel(XmlNode start, XmlNode level, string direction)
	{	XmlDocument input = new XmlDocument();
		XmlDocument output = start.OwnerDocument;
		XmlNodeList subnodes = start.SelectNodes(direction);
		for (int j = 0; j < subnodes.Count; j++)
		{	string ident = subnodes[j].SelectSingleNode("identifier").InnerText;
			if (level.SelectSingleNode("alice[identifier='" + ident + "']") == null)
			{	input.Load(parameters.get("dblink") + "\\" + ident + ".xml");			
				XmlNode type = input.DocumentElement.SelectSingleNode("type");
				if (parameters.get("filter") == "show" || type == null || type.InnerText == "")
				{	
					if (level.ChildNodes.Count%2 == 0) 
						level.AppendChild(output.CreateElement("alice")).InnerXml = input.DocumentElement.InnerXml;
					else level.PrependChild(output.CreateElement("alice")).InnerXml = input.DocumentElement.InnerXml;
				}
			}
		}
	}

	private int findMatch(XmlNode component, XmlNodeList targets)
	{	for (int i = 0; i < targets.Count; i++)
			if (component.SelectSingleNode("identifier").InnerText == targets[i].SelectSingleNode("identifier").InnerText)
				return i;
		return -1;
	}
}





internal class Parameters
{	XmlNode parameters;
	string names = "physical-logical-width-height-user-database-action-param1-param2-dummy";
	FileStream locker;

	internal Parameters(string values)
	{	string newvalues = split("", values, 4);
		lockdirectory(names, newvalues);
		parameters = processActionParameter(formParameters(names, newvalues));
		save();
	}

	internal void diagnose(XmlNode content, string name)
	{	XmlDocument document = new XmlDocument();
		document.AppendChild(document.CreateElement("diagnostics"));
		document.DocumentElement.InnerXml = content.InnerXml;
		document.Save(get("physical") + "\\" + name);
	}
	
	internal void diagnose(string name)
	{	XmlDocument document = new XmlDocument();
		document.AppendChild(document.CreateElement("diagnostics"));
		document.DocumentElement.InnerXml = parameters.InnerXml;
		document.Save(get("physical") + "\\" + name);
	}
	
	internal XmlNode get()
	{	return parameters; 
	}

	internal string get(string name)
	{	if (name == "dblink") return formPhysical(get("physical"), get("logical"), get("database"));
		if (name == "entry")
		{	string reply = get("logical");  
			reply = reply.Substring(reply.LastIndexOf("posd") + 4);
			reply = reply.Substring(0, reply.LastIndexOf("."));
			return reply;
		}
		if (parameters.SelectSingleNode(name) == null) return "";
		else return parameters.SelectSingleNode(name).InnerText; 
	}

	internal XmlDocument load()
	{	return load(get("dblink"), get("user"));
	}	
	
	internal void release()
	{	unlockdirectory();	
	}	
	
	private void lockdirectory(string names, string values)
	{	string address = formPhysical(getM("physical", names, values, "-"), getM("logical", names, values, "-"),
			getM("database", names, values, "-")) + "\\lock.xml";
		locker = new FileStream(address, FileMode.Open);
		try
		{	locker.Lock(0, 1);
		}
		catch (Exception e)
		{	Exception dummy = e;
			lockdirectory(names, values);
		}
	}	
	
	private void unlockdirectory()
	{	locker.Close();
	}	
	
	private XmlDocument load(string address, string user)
	{	XmlDocument reply = new XmlDocument();
		try 
		{	reply.Load(address + "\\parameters" + user + ".xml");
		}
		catch (Exception e) 
		{	Exception dummy = e;
			reply.Load(address + "\\parameters.xml");
		}
		return reply;		
	}	
	
	private void save()
	{	save(parameters.OwnerDocument, get("dblink"), get("user"));
	}	
	
	private void save(XmlDocument document, string address, string user)
	{	document.Save(address + "\\parameters" + user + ".xml");
	}	
	
	private string formPhysical(string physical, string logical, string database)
	{	if ((database.IndexOf("\\") != -1 || database.IndexOf("%") != -1) && logical.IndexOf("localhost") != -1)
			return database;
		else 
			return physical + "\\" + database;
	}
	
	internal Parameters put(string name, string value)
	{	parameters.SelectSingleNode(name).InnerText = value; 
		save();
		return this;
	}
	
	private string split(string front, string back, int position)
	{	string backfront = back.Substring(0, back.IndexOf("-"));
		string backback = back.Substring(back.IndexOf("-") + 1);
		if (position != 0) return split(front + "-" + backfront, backback, position - 1); 
		if (backfront.IndexOf("|") == -1) return front + "--" + back;
		return front + "-" + backfront.Replace("|", "-") + "-" + backback; 
	}

	private XmlNode formParameters(string names, string values)
	{	string database = getM("database", names, values, "-");
		if (getM("action", names, values, "-") == "gotodb") 
			if (getM("param1", names, values, "-") != "null") database = goToDatabase(names, values);
		string address 
					= formPhysical(getM("physical", names, values, "-"), getM("logical", names, values, "-"), database);
		XmlDocument document = load(address, getM("user", names, values, "-"));
		putM(document.DocumentElement, names, values, "-");
		if (getM("action", names, values, "-") == "gotodb") 
		{	put(document.DocumentElement, "database", database);
			put(document.DocumentElement, "action", "|");
		}
		return document.DocumentElement;		
	}
	
	private XmlNode processActionParameter(XmlNode parameters)
	{	string action = get(parameters, "action");
		string param1 = get(parameters, "param1");
		string param2 = get(parameters, "param2");
		if (action == "set") put(parameters, param1, param2);
		else if (action == "sets") putM(parameters, param1, param2, "|");
		else return parameters;
		putM(parameters, "action-param1-param2", "|", "-");
		return parameters;	
	}	

	private string getM(string name, string names, string values, string separator)
	{	int n = names.LastIndexOf(separator);
		int v = values.LastIndexOf(separator);
		if (names.Substring(n + 1) == name) return values.Substring(v + 1);
		else return getM(name, names.Substring(0, n), values.Substring(0, v), separator); 	
	}

	private void putM(XmlNode parameters, string names, string values, string separator)
	{	int n = names.LastIndexOf(separator);
		int v = values.LastIndexOf(separator);
		parameters.SelectSingleNode(names.Substring(n + 1)).InnerText = values.Substring(v + 1);
		if (n != -1) 
		{	if (v != -1) putM(parameters, names.Substring(0, n), values.Substring(0, v), separator);
			else putM(parameters, names.Substring(0, n), values, separator);
		}
	}

	private string get(XmlNode parameters, string name)
	{	if (parameters.SelectSingleNode(name) == null) return "";
		else return parameters.SelectSingleNode(name).InnerText; 
	}

	private void put(XmlNode parameters, string name, string value)
	{	parameters.SelectSingleNode(name).InnerText = value; 
	}
		
	private string goToDatabase(string names, string values)
	{	if (! directoryExists(names, values)) createDatabase(names, values);
		return getM("param1", names, values, "-");
	}	
	
	private bool directoryExists(string names, string values)
	{	DirectoryInfo[] dirs = new DirectoryInfo(getM("physical", names, values, "-")).GetDirectories();
		for (int i = 0; i < dirs.Length; i++) if (dirs[i].Name == getM("param1", names, values, "-")) return true;
		return false;
	}

	private void createDatabase(string names, string values)
	{	Directory.CreateDirectory(getM("physical", names, values, "-") + "\\" + getM("param1", names, values, "-"));	
		string address = formPhysical(getM("physical", names, values, "-"), getM("logical", names, values, "-"),
			getM("database", names, values, "-"));
		XmlDocument document = load(address, "");
		document.DocumentElement.SelectSingleNode("current").InnerText = "s0";
		document.DocumentElement.SelectSingleNode("saved").InnerText = "s0";
		document.DocumentElement.SelectSingleNode("selected").InnerText = "s0";
		document.DocumentElement.SelectSingleNode("left").InnerText = "";
		document.DocumentElement.SelectSingleNode("right").InnerText = "";
		address = getM("physical", names, values, "-") + "\\" + getM("param1", names, values, "-");
		document.Save(address + "\\parameters.xml");
		document.DocumentElement.InnerText = "";
		document.DocumentElement.AppendChild(document.CreateElement("identifier")).InnerText = "s0";
		document.Save(address + "\\s0.xml");
		document.DocumentElement.InnerText = "";
		document.DocumentElement.AppendChild(document.CreateElement("augusta"))
																		.AppendChild(document.CreateElement("alice"));
		document.Save(address + "\\lock.xml");
	}
}


